/*
 * Copyright (c) 2019-2020 Ariadne Conill
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * This software is provided 'as is' and without any warranty, express or
 * implied.  In no event shall the authors be liable for any damages arising
 * from the use of this software.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <linux/if.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>

#ifdef DEBUG
# define DPRINTF(...) fprintf(stderr, __VA_ARGS__)
#else
# define DPRINTF(...)
#endif

#define PROC_NET_DEV "/proc/net/dev"
#define BUFSIZE 4096
#define MAX_MAC_LEN 16

static int kern_socket = -1;
static bool use_config = false;

struct if_definition {
	unsigned char if_macaddr[MAX_MAC_LEN];
	char if_name[IFNAMSIZ+1];
};

static struct if_definition *definitions = NULL;
static size_t definitions_cnt = 0;

static struct if_definition *live_if = NULL;
static size_t live_if_cnt = 0;

static void process_definition(unsigned char if_macaddr[MAX_MAC_LEN],
                               char if_name[IFNAMSIZ + 1])
{
	definitions = realloc(definitions, sizeof(*definitions) * (definitions_cnt + 1));
	if (definitions == NULL)
		errx(1, "failed to allocate more memory for definitions: %s", strerror(errno));

	struct if_definition *def = &definitions[definitions_cnt];

	memset(def, '\0', sizeof *def);

	memcpy(def->if_macaddr, if_macaddr, sizeof def->if_macaddr);
	memcpy(def->if_name, if_name, sizeof def->if_name);

	DPRINTF("Added definition for \"%s\".\n", def->if_name);

	definitions_cnt++;
}

static struct if_definition *find_definition(unsigned char if_macaddr[MAX_MAC_LEN])
{
	size_t iter = 0;
	struct if_definition *if_p;

	for (iter = 0, if_p = &definitions[iter]; iter < definitions_cnt; if_p = &definitions[++iter])
	{
		if (!memcmp(if_p->if_macaddr, if_macaddr, MAX_MAC_LEN))
			return if_p;
	}

	return NULL;
}

static void process_if(unsigned char if_macaddr[MAX_MAC_LEN],
                               char if_name[IFNAMSIZ + 1])
{
	live_if = realloc(live_if, sizeof(*live_if) * (live_if_cnt + 1));
	if (live_if == NULL)
		errx(1, "failed to allocate more memory for live interfaces: %s", strerror(errno));

	struct if_definition *def = &live_if[live_if_cnt];

	memset(def, '\0', sizeof *def);

	memcpy(def->if_macaddr, if_macaddr, sizeof def->if_macaddr);
	memcpy(def->if_name, if_name, sizeof def->if_name);

	DPRINTF("Added \"%s\".\n", def->if_name);

	live_if_cnt++;
}

/*
 * Parse a MAC address into an unsigned char buffer.
 */
static void parse_macaddr(const char *src, unsigned char *buf, size_t size)
{
	const char *base = src;

	memset(buf, '\0', size);

	while (*src && (size_t)(src - base) < size)
	{
		if (*src == ':' || *src == '.')
		{
			src++;
			continue;
		}

		unsigned int templ, temph;
		int count = sscanf(src, "%1x%1x", &temph, &templ);
		if (count != 2)
			break;

		*buf++ = (unsigned char) ((temph << 4) | templ);

		src += 2;
	}
}

/*
 * Parse a definition line.
 * A definition line looks like:
 *      internal     0011.2233.4455
 *
 * The first part is a label (ifname), followed by whitespace,
 * then finally a MAC address.  A MAC address consists of hex
 * digits, colons and periods are ignored.
 */
static void parse_definition_line(char *line)
{
	struct if_definition scratch;
	char *p;

	if (!*line || *line == '#')
		return;

	p = strpbrk(line, "\n");
	*p = 0;

	DPRINTF("Parsing \"%s\".\n", line);

	p = strpbrk(line, " \t");
	*p++ = 0;

	strlcpy(scratch.if_name, line, sizeof scratch.if_name);

	DPRINTF("Name: \"%s\"\n", scratch.if_name);
	DPRINTF("MAC: \"%s\"\n", p);

	parse_macaddr(p, scratch.if_macaddr, sizeof scratch.if_macaddr);

	process_definition(scratch.if_macaddr, scratch.if_name);
}

/*
 * Parse a config file.
 */
static void parse_config_file(const char *filename)
{
	char buf[BUFSIZE];
	FILE *f = fopen(filename, "r");

	if (f == NULL)
		errx(1, "opening config file (%s) failed: %s", filename,
		     strerror(errno));

	DPRINTF("Opened configuration file \"%s\".\n", filename);

	while (fgets(buf, sizeof buf, f) != NULL)
		parse_definition_line(buf);

	fclose(f);
}

#ifdef DEBUG
static void dump_configuration(void)
{
	if (definitions_cnt == 0)
		return;

	DPRINTF("Configuration:\n");

	struct if_definition *iter;
	size_t i = 0;

	for (iter = &definitions[0]; i < definitions_cnt; iter = &definitions[++i])
	{
		DPRINTF("- Interface \"%s\", MAC \"%2x%2x.%2x%2x.%2x%2x\"\n",
			iter->if_name, iter->if_macaddr[0], iter->if_macaddr[1],
			iter->if_macaddr[2], iter->if_macaddr[3], iter->if_macaddr[4],
			iter->if_macaddr[5]);
	}
}
#endif

static void usage(void)
{
	fprintf(stderr, "usage: macifrename [-h] [config-file]\n");
	exit(1);
}

static void foreach_interface(void (*func_p)(const char *ifname))
{
	FILE *fp;
	char buf[BUFSIZE];

	assert(func_p != NULL);

	fp = fopen(PROC_NET_DEV, "r");
	if (fp == NULL)
		errx(1, "unable to open %s: %s", PROC_NET_DEV, strerror(errno));

	while (fgets(buf, sizeof buf, fp) != NULL)
	{
		/* skip header lines */
		if (strchr(buf, '|') != NULL)
			continue;

		char *end_p = strchr(buf, ':');
		if (end_p == NULL)
			continue;

		*end_p = '\0';

		const char *start_p = buf;
		while (*start_p == ' ')
			start_p++;

		func_p(start_p);
	}

	fclose(fp);
}

static int compare_if(const void *a, const void *b)
{
	const struct if_definition *if_a = a, *if_b = b;
	uint64_t mac_a, mac_b;
	const unsigned char *ma_a, *ma_b;

	ma_a = if_a->if_macaddr;
	ma_b = if_b->if_macaddr;

	mac_a = ((uint64_t) ma_a[0] << 40 |
		 (uint64_t) ma_a[1] << 32 |
		 (uint64_t) ma_a[2] << 24 |
		 (uint64_t) ma_a[3] << 16 |
		 (uint64_t) ma_a[4] << 8 |
		 (uint64_t) ma_a[5]);
	mac_b = ((uint64_t) ma_b[0] << 40 |
		 (uint64_t) ma_b[1] << 32 |
		 (uint64_t) ma_b[2] << 24 |
		 (uint64_t) ma_b[3] << 16 |
		 (uint64_t) ma_b[4] << 8 |
		 (uint64_t) ma_b[5]);

	return (int) (mac_a - mac_b);
}

static void discover_if(const char *ifname)
{
	struct ifreq req;

	DPRINTF("Discovering interface %s.\n", ifname);

	memset(&req, '\0', sizeof req);
	strlcpy(req.ifr_name, ifname, sizeof req.ifr_name);

	if (ioctl(kern_socket, SIOCGIFHWADDR, &req) == -1)
		errx(1, "failed to discover interface %s: %s", ifname, strerror(errno));

	process_if((unsigned char *) req.ifr_hwaddr.sa_data, req.ifr_name);
}

static void maybe_take_if_down(const char *ifname, bool *restore)
{
	struct ifreq req;

	strlcpy(req.ifr_name, ifname, sizeof req.ifr_name);

	if (ioctl(kern_socket, SIOCGIFFLAGS, &req) == -1)
		fprintf(stderr, "failed to take down %s: %s\n", ifname, strerror(errno));

	if (req.ifr_flags & IFF_UP)
		*restore = true;

	req.ifr_flags &= ~IFF_UP;

	if (ioctl(kern_socket, SIOCSIFFLAGS, &req) == -1)
		fprintf(stderr, "failed to take down %s: %s\n", ifname, strerror(errno));
}

static void restore_if(const char *ifname)
{
	struct ifreq req;

	strlcpy(req.ifr_name, ifname, sizeof req.ifr_name);
	req.ifr_flags = IFF_UP;

	if (ioctl(kern_socket, SIOCSIFFLAGS, &req) == -1)
		fprintf(stderr, "failed to restore %s: %s\n", ifname, strerror(errno));
}

static void rename_interface(const char *old, const char *new)
{
	struct ifreq req;
	bool restore = false;

	strlcpy(req.ifr_name, old, sizeof req.ifr_name);
	strlcpy(req.ifr_newname, new, sizeof req.ifr_newname);

	maybe_take_if_down(old, &restore);

	if (ioctl(kern_socket, SIOCSIFNAME, &req) == -1)
		errx(1, "failed to rename interface %s to %s: %s", old, new,
		     strerror(errno));

	if (restore)
		restore_if(req.ifr_newname);
}

static void rename_interfaces_simple(void)
{
	size_t cnt = 0, iter = 0;
	char buf[BUFSIZE];
	struct if_definition *if_p;

	/* first rename the interfaces with _t suffix. */
	for (iter = 0, if_p = &live_if[iter]; iter < live_if_cnt; if_p = &live_if[++iter])
	{
		if (strncmp(if_p->if_name, "eth", 3))
			continue;

		snprintf(buf, sizeof buf, "%s_t", if_p->if_name);
		DPRINTF("Temporarily rename %s to %s.\n", if_p->if_name, buf);

		rename_interface(if_p->if_name, buf);
		strlcpy(if_p->if_name, buf, sizeof(if_p->if_name));
	}

	/* now rename the interfaces with _t suffix to their permanent names */
	for (iter = 0, if_p = &live_if[iter]; iter < live_if_cnt; if_p = &live_if[++iter])
	{
		if (strncmp(if_p->if_name, "eth", 3))
			continue;

		snprintf(buf, sizeof buf, "eth%zu", cnt);
		DPRINTF("Permanently rename %s to %s.\n", if_p->if_name, buf);

		rename_interface(if_p->if_name, buf);
		cnt++;
	}
}

static void rename_interfaces_config(void)
{
	size_t iter = 0;
	struct if_definition *if_p;

	for (iter = 0, if_p = &live_if[iter]; iter < live_if_cnt; if_p = &live_if[++iter])
	{
		DPRINTF("Looking for definition matching %s.\n", if_p->if_name);

		struct if_definition *ifd = find_definition(if_p->if_macaddr);
		if (ifd == NULL)
			continue;

		DPRINTF("Permanently rename %s to %s.\n", if_p->if_name, ifd->if_name);
		rename_interface(if_p->if_name, ifd->if_name);
	}
}

int main(int argc, const char *argv[])
{
	if (argc > 1 && strcmp(argv[1], "-h"))
	{
		DPRINTF("Loading configuration.\n");

		parse_config_file(argv[1]);
#ifdef DEBUG
		dump_configuration();
#endif

		use_config = true;
	}
	else if (argc > 1 && !strcmp(argv[1], "-h"))
		usage();

	DPRINTF("Opening kernel socket.\n");
	kern_socket = socket(AF_INET, SOCK_DGRAM, 0);
	if (kern_socket < 0)
		errx(1, "unable to open kernel networking socket: %s",
		     strerror(errno));

	DPRINTF("Discovering interfaces.\n");
	foreach_interface(discover_if);

	if (live_if_cnt == 0)
		return EXIT_SUCCESS;

	DPRINTF("Sorting live interfaces.\n");
	qsort(live_if, live_if_cnt, sizeof *live_if, compare_if);

	DPRINTF("Applying configuration.\n");
	if (!use_config)
		rename_interfaces_simple();
	else
		rename_interfaces_config();

	DPRINTF("Closing kernel socket.\n");
	close(kern_socket);

	return 0;
}
